package bzr4idea;

import bzr4idea.util.Assert;
import bzr4idea.util.IOHelper;
import bzr4idea.util.Log;
import bzr4idea.util.IdeaHelper;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vcs.FilePath;
import com.intellij.vcsUtil.VcsUtil;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.vcs.bazaar.client.*;
import org.vcs.bazaar.client.commandline.commands.options.Option;
import org.vcs.bazaar.client.commandline.syntax.*;
import org.vcs.bazaar.client.core.BazaarClientException;
import org.vcs.bazaar.client.core.BranchLocation;
import org.vcs.bazaar.client.integration.CygwinFilePathAdapter;
import org.vcs.bazaar.client.utils.BazaarUtilities;

import java.io.*;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.WritableByteChannel;
import java.util.*;

import git4idea.config.BzrVcsSettings;

/**
 * @author Patrik Beno
 */
public class Bzr {

  static public final Option[] EMPTY_OPTIONS = new Option[0];

  @NonNls
  static public final String ADMIN_DIR = ".bzr";

  VirtualFile root;

  IBazaarClient bzr;

  static public void initBazaarClient(BzrVcsSettings settings) throws BzrException {
    try {
      BazaarClientPreferences prefs = BazaarClientPreferences.getInstance();
      prefs.set(BazaarPreference.EXECUTABLE, settings.GIT_EXECUTABLE);
      if (isWindows() && settings.cygwin) {
        prefs.set(BazaarPreference.FILE_PATH_ADAPTER, CygwinFilePathAdapter.class.getName());
      }
      BazaarClientFactory.setupBestAvailableBackend(true);

      IBazaarClient client = BazaarClientFactory.createClient(BazaarClientFactory.getPreferredClientType());
      if (client == null) {
        throw new BzrException(null, "Cannot initialize bzr client");
      }

      Log.info("BazaarClientFactory initialized.");
    } catch (BazaarClientException e) {
      throw new BzrException(e, "Cannot initialize bzr client");
    }
  }

  static public boolean isWindows() {
    return System.getProperty("os.name").contains("Windows");
  }

  static public Bzr forBzrRoot(VirtualFile vfile) {
    return new Bzr(vfile);
  }

  Bzr(VirtualFile root) {
    try {
      this.root = root;
      bzr = BazaarClientFactory.createClient(BazaarClientFactory.getPreferredClientType());
      bzr.setWorkDir(IdeaHelper.getIOFile(root));
    } catch (BazaarClientException e) {
      throw Assert.notYetHandled(e);
    }
  }

  public IBazaarClient getClient() {
    return bzr;
  }

  public List<IBazaarItemInfo> list(File dir, Option... options) throws BzrException {
    try {
      IBazaarItemInfo[] items = bzr.ls(dir, null, options);
      return Arrays.asList(items);

    } catch (BazaarClientException e) {
      throw bzrexception(e);
    }
  }

  /**
   * Gets Bazaar tree status
   * @param vf virtual path of tree root to get status for, or null to get status of the current working dir
   * @return Bazaar tree status
   */
  @NotNull
  public BazaarTreeStatus status(@Nullable VirtualFile vf) throws BzrException {
    final File f = vf != null ? IdeaHelper.getIOFile( vf ) : null;
    try {
      BazaarTreeStatus status = bzr.status( f != null ? new File[]{ f } : null );
      return status;
    } catch (BazaarClientException e) {
      if ( f != null && !f.exists()) { // hacking http://jira.greenhorn.sk/browse/BZR-31
        BazaarTreeStatus status = new BazaarTreeStatus(
            new HashSet<IBazaarStatus>(Arrays.asList(BazaarStatus.create(f, IdeaHelper.getIOFile( root ), BazaarStatusKind.NONE))),
            Collections.<IBazaarLogMessage>emptyList()
        );
        return status;
      }
      throw bzrexception(e);
    }
  }

  public IBazaarAnnotation annotate(@NotNull FilePath path, @Nullable IBazaarRevisionSpec revSpec) throws BzrException {
    final File f = path.getIOFile();
    try {
      return revSpec != null ? bzr.annotate( f, IAnnotateOptions.REVISION.withArgument( revSpec.toString() ) )
                         : bzr.annotate( f );
    }
    catch (BazaarClientException e ) {
      throw new BzrException( e, "Failed to annotate file '%s'", path.getPresentableUrl() );
    }
  }

  public IBazaarInfo info() throws BzrException {
    try {
      return bzr.info( IdeaHelper.getIOFile(root) );
    } catch (BazaarClientException e) {
      throw new BzrException( e, "Failed to get info for '%s'", root.getPresentableName() );
    }
  }

  public List<IBazaarLogMessage> log(@NotNull FilePath path, Collection<Option> options ) throws BzrException {
    final File f = path.getIOFile();
    try {
      return bzr.log( f, options.toArray( EMPTY_OPTIONS ) );
    }
    catch (BazaarClientException e ) {
      throw new BzrException( e, "Failed to get log for file '%s'", path.getPresentableUrl() );
    }
  }

  public List<IBazaarLogMessage> log(@NotNull FilePath path, @Nullable IBazaarRevisionSpec revSpec, @Nullable Integer limit) throws BzrException {
    List<Option> options = new ArrayList<Option>();
    if ( revSpec != null ) {
      options.add( ILogOptions.REVISION.withArgument( revSpec.toString() ) );
    }
    if ( limit != null ) {
      options.add( ILogOptions.LIMIT.withArgument( limit.toString() ) );
    }
    return log( path, options );
  }

  public void add(File f) throws BzrException {
    try {
      bzr.add(new File[]{f}, IAddOptions.NO_RECURSE);
    } catch (BazaarClientException e) {
      throw bzrexception(e);
    }
  }

  public void remove(List<FilePath> files) throws BzrException {
    try {
      bzr.remove(IdeaHelper.toFileArray(files), IRemoveOptions.KEEP);
    } catch (BazaarClientException e) {
      throw bzrexception(e);
    }
  }

  public void move(FilePath src, VirtualFile dst) throws BzrException {
    try {
      bzr.move(src.getIOFile(), IdeaHelper.getIOFile(dst), IMoveOptions.AFTER);
    } catch (BazaarClientException e) {
      throw bzrexception(e);
    }
  }

  public void revert(List<FilePath> files) throws BzrException {
    try {
      bzr.revert(IdeaHelper.toFileArray(files), IRevertOptions.NO_BACKUP);

    } catch (BazaarClientException e) {
      throw bzrexception(e);
    }
  }

  public BazaarRevision commit(String comment, File... files) throws BzrException {
    try {
      File root = BazaarUtilities.getRootBranch(files[0]);
      bzr.commit(files, comment);
      BazaarRevision revno = revno(root);
      return revno;

    } catch (BazaarClientException e) {
      throw bzrexception(e);
    }
  }

  public void push( BranchLocation dst, @Nullable Collection<Option> options ) throws BzrException {
    try {
      bzr.push( dst, options != null ? options.toArray(EMPTY_OPTIONS) : EMPTY_OPTIONS );
    } catch (BazaarClientException e) {
      throw bzrexception(e);
    }
  }

  public void pull( BranchLocation src, @Nullable Collection<Option> options ) throws BzrException {
    try {
      bzr.pull( src, options != null ? options.toArray(EMPTY_OPTIONS) : EMPTY_OPTIONS );
    } catch (BazaarClientException e) {
      throw bzrexception(e);
    }
  }

  public void commit(String message, Set<FilePath> files) throws BzrException {
    try {
      List<File> ffiles = new ArrayList<File>();
      for (FilePath f : files) {
        ffiles.add(f.getIOFile());
      }

      bzr.commit(ffiles.toArray(new File[ffiles.size()]), message);

    } catch (BazaarClientException e) {
      throw bzrexception(e);
    }
  }


  public void cat(OutputStream out, File file,
                         BazaarRevision revision) throws IOException, BzrException {
    InputStream input = null;
    try {
      input = bzr.cat(file, revision);

      // get an channel from the outout
      final ReadableByteChannel inputChannel = Channels.newChannel(input);
      final WritableByteChannel outputChannel = Channels.newChannel(out);
      try {
        // from ChannelTools.fastChannelCopy()
        final ByteBuffer ioBuffer = ByteBuffer.allocateDirect(16 * 1024);
        while (inputChannel.read(ioBuffer) != -1) {
          // prepare the buffer to be drained
          ioBuffer.flip();
          outputChannel.write(ioBuffer);
          // If partial transfer, shift remainder down
          // If buffer is empty, same as doing clear()
          ioBuffer.compact();
        }
        // EOF will leave buffer in fill state
        ioBuffer.flip();
        // make sure the buffer is fully drained.
        while (ioBuffer.hasRemaining()) {
          outputChannel.write(ioBuffer);
        }

      } finally {
        IOHelper.close(inputChannel);
        IOHelper.close(outputChannel);
      }

    } catch (BazaarClientException e) {
      throw bzrexception(e);

    } finally {
      IOHelper.close(input);
    }
  }

  public void cat(OutputStream out, FilePath path,
                         BazaarRevision revision) throws IOException, BzrException {
    cat( out, path.getIOFile(), revision );
  }

   /**
   * Checks if the given status contains *any* of the kinds listed
   * @param status
   * @param kinds
   * @return
   */
  public boolean contains(BazaarTreeStatus status, BazaarStatusKind ... kinds) {
    for (BazaarStatusKind kind : kinds) {
      if (contains(status, kind)) {
        return true;
      }
    }
    return false;
  }

  public boolean contains(BazaarTreeStatus status, BazaarStatusKind kind) {
    for (IBazaarStatus st : status.getStatus()) {
      if (st.contains(kind)) {
        return true;
      }
    }
    return false;
  }

  static public VirtualFile root(VirtualFile vf) {
    File root = BazaarUtilities.getRootBranch(IdeaHelper.getIOFile(vf));
    return VcsUtil.getVirtualFile(root);
  }

  public BazaarRevision revno(File file) throws BzrException {
    try {
      BazaarRevision rev = bzr.revno(new BranchLocation("."));
      return rev;
    } catch (BazaarClientException e) {
      throw bzrexception(e);
    }
  }

  BzrException bzrexception(Throwable t) {
    return new BzrException(t, "BZR error");
  }

}
