package git4idea;
/*
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for
 * the specific language governing permissions and limitations under the License.
 *
 * Copyright 2007 Decentrix Inc
 * Copyright 2007 Aspiro AS
 * Copyright 2008 MQSoftware
 * Authors: gevession, Erlend Simonsen & Mark Scott
 *
 * This code was originally derived from the MKS & Mercurial IDEA VCS plugins
 */

import com.intellij.openapi.project.Project;
import com.intellij.openapi.vcs.VcsException;
import com.intellij.openapi.vcs.history.VcsRevisionNumber;
import com.intellij.openapi.vfs.VirtualFile;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.client.IBazaarLogMessage;

import java.util.Arrays;
import java.util.Date;
import java.util.List;

import bzr4idea.Bzr;
import bzr4idea.util.IdeaHelper;

/**
 * Git revision number
 */
public class GitRevisionNumber implements VcsRevisionNumber {
  /**
   * the name of tip revision
   */
  @NonNls public static final String TIP = "HEAD";
  /**
   * the hash from 40 zeros representing not yet created commit
   */
  public static final String NOT_COMMITED_HASH;

  static {
    char[] data = new char[40];
    Arrays.fill(data, '0');
    NOT_COMMITED_HASH = new String(data);
  }

  /**
   * bazaar revision container
   */
  @NotNull private final BazaarRevision myRevision;
  /**
   * the revision number (40 character hashcode, tag, or reference). In some cases incomplete hashcode could be used.
   */
  /* git4idea
  @NotNull private final String myRevisionStr;
  */
  /**
   * the date when revision created
   */
  @NotNull private final Date myTimestamp;


  /**
   * A constrctuctor for TIP revision
   */
  public GitRevisionNumber() {
    // TODO review usages
    /* git4idea
    myRevisionStr = TIP;
    */
    myRevision = BazaarRevision.TIP;
    myTimestamp = new Date();
  }

  /**
   * A constructor from version. The current date is used.
   *
   * @param version the version number.
   */
  public GitRevisionNumber(@NonNls @NotNull String version) {
    // TODO review usages
    /* git4idea
    myRevisionStr = version;
    */
    myRevision = BazaarRevision.getRevision( BazaarRevision.Prefix.REVNO, version );
    myTimestamp = new Date();
  }

  /**
   * A consctructor from version and time
   *
   * @param version   the version number
   * @param timeStamp the time when the version has been created
   */
  public GitRevisionNumber(@NotNull String version, @NotNull Date timeStamp) {
    myTimestamp = timeStamp;
    /* git4idea
    myRevisionStr = version;
    */
    myRevision = BazaarRevision.getRevision( BazaarRevision.Prefix.REVNO, version );
  }

  /**
   * A consctructor from bzr revision
   *
   * @param revision bzr revision
   */
  public GitRevisionNumber(@NotNull BazaarRevision revision) {
    myRevision = revision;
    // TODO: initialize timestamp?
    myTimestamp = new Date();
  }

  /**
   * A consctructor from bzr revision and timestamp
   *
   * @param revision bzr revision
   */
  public GitRevisionNumber(@NotNull BazaarRevision revision, @NotNull Date timestamp) {
    myRevision = revision;
    myTimestamp = timestamp;
  }

  /**
   * {@inheritDoc}
   *
   * @see #getRev()
   */
  @NotNull
  public String asString() {
    /* git4idea
    return myRevisionStr;
    */
    return myRevision.toString();
  }

  /**
   * @return revision time
   */
  @NotNull
  public Date getTimestamp() {
    return myTimestamp;
  }

  /**
   * @return revision number
   */
  @NotNull
  public String getRev() {
    /* git4idea
    return myRevisionStr;
    */
    return myRevision.toString();
  }

  /**
   * @return the short revision number. The revision number likely unambiguously indentify local revision, however in rare cases there could be conflicts.
   */
  @NotNull
  public String getShortRev() {
    /* git4idea
    if (myRevisionStr.length() == 0) return "";
    if (myRevisionStr.length() == 40) return myRevisionStr.substring(0, 8);
    if (myRevisionStr.length() > 40)  // revision string encoded with date too
    {
      return myRevisionStr.substring(myRevisionStr.indexOf("[") + 1, 8);
    }
    return myRevisionStr;
    */
    // TODO short version of bazaar revision?
    return myRevision.toString();
  }

  /**
   * @return the BazaarRevision object wrapped inside
   */
  @NotNull
  public BazaarRevision getBzrRevision() {
    return myRevision;
  }

  /**
   * {@inheritDoc}
   */
  public int compareTo(VcsRevisionNumber crev) {
    if (this == crev) return 0;

    if (crev instanceof GitRevisionNumber) {
      GitRevisionNumber crevg = (GitRevisionNumber)crev;
      /* git4idea
      if ((crevg.myRevisionStr != null) && myRevisionStr.equals(crevg.myRevisionStr)) {
        return 0;
      }
      */
      if ((crevg.myRevision != null) && myRevision.equals(crevg.myRevision)) {
        return 0;
      }

      /* git4idea
      if ((crevg.myRevisionStr.indexOf("[") > 0) && (crevg.myTimestamp != null)) {
        return myTimestamp.compareTo(crevg.myTimestamp);
      }

      // check for parent revs
      String crevName = null;
      String revName = null;
      int crevNum = -1;
      int revNum = -1;

      if (crevg.myRevisionStr.contains("~")) {
        int tildeIdx = crevg.myRevisionStr.indexOf('~');
        crevName = crevg.myRevisionStr.substring(0, tildeIdx);
        crevNum = Integer.parseInt(crevg.myRevisionStr.substring(tildeIdx));
      }

      if (myRevisionStr.contains("~")) {
        int tildeIdx = myRevisionStr.indexOf('~');
        revName = myRevisionStr.substring(0, tildeIdx);
        revNum = Integer.parseInt(myRevisionStr.substring(tildeIdx));
      }

      if (crevName == null && revName == null) {
        return myTimestamp.compareTo(crevg.myTimestamp);
      }
      else if (crevName == null) {
        return 1;  // I am an ancestor of the compared revision
      }
      else if (revName == null) {
        return -1; // the compared revision is my ancestor
      }
      else {
        return revNum - crevNum;  // higher relative rev numbers are older ancestors
      }
      */
      return myRevision.compareTo(crevg.myRevision);
    }

    return -1;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) return true;
    if ((obj == null) || (obj.getClass() != getClass())) return false;

    GitRevisionNumber test = (GitRevisionNumber)obj;
    /* git4idea
    // TODO normailize revision string?
    return myRevisionStr.equals(test.myRevisionStr);
    */
    return myRevision.equals(test.myRevision);
  }

  @Override
  public int hashCode() {
    /* git4idea
    return myRevisionStr.hashCode();
    */
    return myRevision.hashCode();
  }

  /**
   * @return a revision string that refers to the parent revision relatively
   *         to the current one. The git operator "~" is used. Note that in case of merges,
   *         the first revision of several will referred.
   */
  public String getParentRevisionStr() {
    /* git4idea
    String rev = myRevisionStr;
    int bracketIdx = rev.indexOf("[");
    if (bracketIdx > 0) {
      rev = myRevisionStr.substring(bracketIdx + 1, myRevisionStr.indexOf("]"));
    }

    int tildeIdx = rev.indexOf("~");
    if (tildeIdx > 0) {
      int n = Integer.parseInt(rev.substring(tildeIdx)) + 1;
      return rev.substring(0, tildeIdx) + "~" + n;
    }
    return rev + "~1";
    */
    final String rev;
    if ( myRevision.getPrefix() == BazaarRevision.Prefix.REVNO ) {
      // extract x from x.y.z revision number, where x is the parent revision
      final int dotIdx = myRevision.getValue().indexOf('.');
      if ( dotIdx > 0 ) {
        rev = myRevision.getValue().substring( 0, dotIdx );
      } else {
        rev = null;
      }
    } else {
      rev = null;
    }
    return rev;
  }

  /**
   * Create revision number from string
   *
   * @param rev a revision string
   * @return revision number (the same as {@link #GitRevisionNumber(String)}
   */
  public static GitRevisionNumber createRevision(String rev) {
    return new GitRevisionNumber(rev);
  }

  /**
   * Create revision number from string
   *
   * @param rev a revision string
   * @param timestamp a revision date
   * @return revision number (the same as {@link #GitRevisionNumber(String)}
   */
  public static GitRevisionNumber createRevision(String rev, Date timestamp) {
    return new GitRevisionNumber(rev, timestamp);
  }

  /**
   * Resolve revision number for the specified revision
   *
   * @param project a project
   * @param vcsRoot a vcs root
   * @param rev     bazaar revision
   * @return a resolved revision number with correct time
   */
  public static GitRevisionNumber resolve(Project project, VirtualFile vcsRoot, @NotNull BazaarRevision rev) throws VcsException {
    /* bzr4idea
    GitSimpleHandler h = new GitSimpleHandler(project, vcsRoot, GitHandler.REV_LIST);
    h.setNoSSH(true);
    h.setSilent(true);
    h.addParameters("--timestamp", "--max-count=1", rev);
    h.endOptions();
    StringTokenizer stk = new StringTokenizer(h.run(), "\n\r \t", false);
    Date timestamp = GitUtil.parseTimestamp(stk.nextToken());
    return new GitRevisionNumber(stk.nextToken(), timestamp);
    */
    Bzr bzr = Bzr.forBzrRoot( vcsRoot );
    List<IBazaarLogMessage> tipLog = bzr.log(IdeaHelper.getFilePath( vcsRoot ), rev, 1);
    return tipLog.size() == 1 ? new GitRevisionNumber( tipLog.get(0).getRevision() ) : null;
  }

  /**
   * Resolve revision number for the specified revision
   *
   * @param project a project
   * @param vcsRoot a vcs root
   * @param rev     a revision expression
   * @return a resolved revision number with correct time
   */
  public static GitRevisionNumber resolve(Project project, VirtualFile vcsRoot, @NonNls String rev) throws VcsException {
    return resolve( project, vcsRoot, BazaarRevision.getRevision(BazaarRevision.Prefix.REVNO, rev ) );
  }
}
