/*
 * Copyright 2000-2008 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package git4idea.actions;

import git4idea.i18n.GitBundle;
import git4idea.rebase.GitRebaseUtils;
import git4idea.rebase.GitRebaseActionDialog;
import git4idea.commands.GitSimpleHandler;
import git4idea.commands.GitHandler;
import git4idea.commands.GitHandlerUtil;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vcs.VcsException;
import com.intellij.openapi.ui.Messages;

import java.util.List;
import java.util.Set;
import java.util.Iterator;

/**
 * Rebase abort action
 */
public class GitRebaseAbort extends GitRepositoryAction {

  /**
   * {@inheritDoc}
   */
  protected void perform(@NotNull Project project,
                         @NotNull List<VirtualFile> gitRoots,
                         @NotNull VirtualFile defaultRoot,
                         Set<VirtualFile> affectedRoots,
                         List<VcsException> exceptions) throws VcsException {

    Messages.showErrorDialog("Not yet implemented", "Error");
    if (true) return;
    
    // remote all roots where there are no rebase in progress
    for (Iterator<VirtualFile> i = gitRoots.iterator(); i.hasNext();) {
      if (!GitRebaseUtils.isRebaseInTheProgress(i.next())) {
        i.remove();
      }
    }
    if (gitRoots.size() == 0) {
      Messages.showErrorDialog(project, GitBundle.getString("rebase.action.no.root"), GitBundle.getString("rebase.action.error"));
      return;
    }
    final VirtualFile root;
    if (gitRoots.size() == 1) {
      root = gitRoots.get(0);
    }
    else {
      if (!gitRoots.contains(defaultRoot)) {
        defaultRoot = gitRoots.get(0);
      }
      GitRebaseActionDialog d = new GitRebaseActionDialog(project, getActionTitle(), gitRoots, defaultRoot);
      root = d.selectRoot();
      if (root == null) {
        return;
      }
    }
    affectedRoots.add(root);
    GitSimpleHandler h = new GitSimpleHandler(project, root, GitHandler.REBASE);
    h.addParameters(getOptionName());
    GitHandlerUtil.doSynchronously(h, getActionTitle(), h.printableCommandLine());
  }

  /**
   * {@inheritDoc}
   */
  @NotNull
  protected String getActionName() {
    return GitBundle.getString("rebase.abort.action.name");
  }


  /**
   * @return the option name
   */
  @NonNls
  protected String getOptionName() {
    return "--abort";
  }

  /**
   * @return title for root selection dialog
   */
  protected String getActionTitle() {
    return GitBundle.getString("rebase.abort.action.name");
  }
}
