/*
 * Copyright 2000-2007 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package git4idea.changes;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.vcs.FileStatus;
import com.intellij.openapi.vcs.VcsException;
import com.intellij.openapi.vcs.FilePath;
import com.intellij.openapi.vcs.changes.Change;
import com.intellij.openapi.vcs.changes.ContentRevision;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.util.CommonProcessors;
import com.intellij.vcsUtil.VcsUtil;
import git4idea.GitContentRevision;
import git4idea.GitRevisionNumber;
import git4idea.GitUtil;
import git4idea.commands.GitHandler;
import git4idea.commands.GitSimpleHandler;

import java.util.*;

import bzr4idea.Bzr;
import org.vcs.bazaar.client.BazaarTreeStatus;
import org.vcs.bazaar.client.IBazaarStatus;
import org.vcs.bazaar.client.BazaarStatusKind;
import org.vcs.bazaar.client.BazaarRevision;
import org.jetbrains.annotations.NotNull;

/**
 * A collector for changes in the Git. It is introduced because changes are not
 * cannot be got as a sum of stateless operations.
 */
class ChangeCollector {
  private static final Logger LOG = Logger.getInstance("bzr4idea.ChangeCollector");

  /**
   * a vcs root for changes
   */
  private final VirtualFile myVcsRoot;
  /**
   * a project for change collection
   */
  private final Project myProject;
  /**
   * Unversioned files
   */
  private final List<VirtualFile> myUnversioned = new ArrayList<VirtualFile>();
  /**
   * Ignored files
   */
  private final List<VirtualFile> myIgnored = new ArrayList<VirtualFile>();
  /**
   * Names that are listed as unmerged
   */
  private final Set<String> myUnmergedNames = new HashSet<String>();
  /**
   * Names that are listed as unmerged
   */
  private final List<Change> myChanges = new ArrayList<Change>();
  /**
   * This flag indicates that collecting changes has been failed.
   */
  private boolean myIsFailed = true;
  /**
   * This flag indicates that collecting changes has been started
   */
  private boolean myIsCollected = false;

  private List<VirtualFile> myRenamed = new ArrayList<VirtualFile>();
  private List<VirtualFile> myCopied = null;
  private Map<String, PathModification> myPrevious = new HashMap<String, PathModification>();

  /**
   * A constructor
   *
   * @param project a project
   * @param vcsRoot a vcs root
   */
  public ChangeCollector(final Project project, final VirtualFile vcsRoot) {
    myVcsRoot = vcsRoot;
    myProject = project;
  }

  /**
   * Get unversioned files
   *
   * @return an unversioned files
   * @throws VcsException if there is a problem with executing Git
   */
  public Collection<VirtualFile> unversioned() throws VcsException {
    ensureCollected();
    return myUnversioned;
  }

  /**
   * Get ignored files
   *
   * @return an ignored files
   * @throws VcsException if there is a problem with executing Git
   */
  public Collection<VirtualFile> ignored() throws VcsException {
    ensureCollected();
    return myIgnored;
  }

  /**
   * Get changes
   *
   * @return an unversioned files
   * @throws VcsException if there is a problem with executing Git
   */
  public Collection<Change> changes() throws VcsException {
    ensureCollected();
    return myChanges;
  }


  /**
   * Ensure that changes has been collected.
   *
   * @throws VcsException an exception
   */
  private void ensureCollected() throws VcsException {
    if (myIsCollected) {
      if (myIsFailed) {
        throw new IllegalStateException("The method should not be called after after exception has been thrown.");
      }
      else {
        return;
      }
    }
    myIsCollected = true;
    /* git4idea
    collectUnmergedAndUnversioned();
    collectDiffChanges();
    */
    collectChanges();
    myIsFailed = false;
  }

  /**
   * Collect diff with head
   *
   * @throws VcsException if there is a problem with running git
   */
  @Deprecated
  private void collectDiffChanges() throws VcsException {
    GitSimpleHandler handler = new GitSimpleHandler(myProject, myVcsRoot, GitHandler.DIFF);
    handler.addParameters("--name-status", "--diff-filter=ADCMRUX", "-M", "HEAD");
    handler.setNoSSH(true);
    handler.setSilent(true);
    handler.setStdoutSuppressed(true);
    handler.endOptions();
    try {
      String output = handler.run();
      String[] lines = output.split("\n");
      GitChangeUtils
        .parseChanges(myProject, myVcsRoot, null, GitChangeUtils.loadRevision(myProject, myVcsRoot, "HEAD"), lines, 0, myChanges,
                      myUnmergedNames);
    }
    catch (VcsException ex) {
      if (!GitChangeUtils.isHeadMissing(ex)) {
        throw ex;
      }
      handler = new GitSimpleHandler(myProject, myVcsRoot, GitHandler.LS_FILES);
      handler.addParameters("--cached");
      handler.setNoSSH(true);
      handler.setSilent(true);
      handler.setStdoutSuppressed(true);
      // During init diff does not works because HEAD
      // will apppear only after the first commit.
      // In that case added files are cached in index.
      String output = handler.run();
      if (output.length() > 0) {
        StringTokenizer tokenizer = new StringTokenizer(output, "\n\r");
        while (tokenizer.hasMoreTokens()) {
          final String s = tokenizer.nextToken();
          Change ch = new Change(null, GitContentRevision.createRevision(myVcsRoot, s, null, myProject, false), FileStatus.ADDED);
          myChanges.add(ch);
        }
      }
    }
  }

  /**
   * Collect unversioned and unmerged files
   *
   * @throws VcsException if there is a problem with running git
   */
  @Deprecated
  private void collectUnmergedAndUnversioned() throws VcsException {
    // prepare handler
    GitSimpleHandler handler = new GitSimpleHandler(myProject, myVcsRoot, GitHandler.LS_FILES);
    handler.addParameters("-v", "--others", "--unmerged", "--exclude-standard");
    handler.setSilent(true);
    handler.setNoSSH(true);
    handler.setStdoutSuppressed(true);
    // run handler and collect changes
    String list = handler.run();
    for (String line : list.split("\n")) {
      if (line.length() == 0) {
        continue;
      }
      String[] tokens = line.split("[\t ]+");
      String file = GitUtil.unescapePath(tokens[tokens.length - 1]);
      if ("?".equals(tokens[0])) {
        myUnversioned.add(myVcsRoot.findFileByRelativePath(file));
      }
      else { //noinspection HardCodedStringLiteral
        if ("M".equals(tokens[0])) {
          if (!myUnmergedNames.add(file)) {
            continue;
          }
          // TODO handle conflict rename-modify
          // TODO handle conflict copy-modify
          // TODO handle conflict delete-modify
          // TODO handle conflict rename-delete
          // assume modify-modify conflict
          ContentRevision before = GitContentRevision.createRevision(myVcsRoot, file, new GitRevisionNumber("orig_head"), myProject, false);
          ContentRevision after = GitContentRevision.createRevision(myVcsRoot, file, null, myProject, false);
          myChanges.add(new Change(before, after, FileStatus.MERGED_WITH_CONFLICTS));
        }
        else {
          throw new VcsException("Unsupported type of the merge conflict detected: " + line);
        }
      }
    }
  }

    /**
     * Collect all diffs with head
     *
     * @throws VcsException if there is a problem with running bzr
     */
  private void collectChanges() throws VcsException {
    Bzr bzr = Bzr.forBzrRoot( myVcsRoot );
    BazaarTreeStatus statuses = bzr.status( myVcsRoot );
    if ( LOG.isDebugEnabled() ) LOG.debug( String.format( "BazaarTreeStatus: %s", statuses.toString() ) );

    /* bzr4idea - loads timestamp
    GitRevisionNumber tip = GitChangeUtils.loadRevision(myProject, myVcsRoot, "HEAD");
    */
    GitRevisionNumber tip = new GitRevisionNumber( BazaarRevision.TIP );
    boolean refreshedDirState = false;
    for (IBazaarStatus itemStatus : statuses.getStatus()) {
      if ( LOG.isDebugEnabled() ) LOG.debug( String.format( "Processing bzr's status record: %s", itemStatus.toString() ) );
      /* svn4idea legacy
      if (context.isCanceled()) {
        throw new ProcessCanceledException();
      }
      if (!refreshedDirState) {
        loadDirstateFile(itemStatus.getBranchRoot());
        refreshedDirState = true;
      }
      */
      // @TODO: support symlinks item kind?
      processFileStatus(null, tip, itemStatus);
    }
  }

  private void recursivelyAddFiles(List<VirtualFile> dest, final VirtualFile vFile) {
    final CommonProcessors.CollectUniquesProcessor<VirtualFile> processor = new CommonProcessors.CollectUniquesProcessor<VirtualFile>();
    VfsUtil.processFilesRecursively(vFile, processor);
    dest.addAll( processor.getResults() );
  }

  private void processFileStatus( final GitRevisionNumber thisRevision, @NotNull final GitRevisionNumber parentRevision,
                                  @NotNull final IBazaarStatus status) throws VcsException {
    // @FIXME (ast) remove, since DVCS(bzr) does not need locking and doesn't support it?
  //  if (filePath.isDirectory() && status.isLocked()) {
  //    context.getBuilder().processLockedFolder(filePath.getVirtualFile());
  //  }
    // @FIXME (ast) remove, since DVCS(bzr) does not support copying
  //    if (status.getContentsStatus() == SVNStatusType.STATUS_ADDED && status.getCopyFromURL() != null) {
  //      context.addPathModification(filePath, status, status.getCopyFromURL());
  //    }
    // @FIXME (ast) remove, since DVCS(bzr) does not support copying
    String path = status.getPath();
    VirtualFile vFile = myVcsRoot.findFileByRelativePath( path );

    PathModification previousPath = getExistingPathModification( path );
    if (status.contains(BazaarStatusKind.RENAMED)) {
      if ( LOG.isDebugEnabled() ) LOG.debug( String.format("File '%s' is renamed/moved. Old path is '%s'", path, status.getPreviousPath() ) );
      previousPath = addPathModification( status, previousPath );
    } else if (previousPath != null) {
        LOG.debug( String.format("File '%s' was %s, previous path is '%s'",
                                 status.getPath(),
                                 previousPath.isCopy() ? "copied" : "renamed",
                                 previousPath.getOriginalPath() ) );
    }
    // @FIXME (ast) remove, it's just to support copy+delete logic
  //    else if (status.contains(BazaarStatusKind.DELETED)) {
  //      if ( LOG.isDebugEnabled() ) LOG.debug( String.format("File '%s' is deleted", filePath.getPath() ) );
  //      context.getDeletedFiles().add(new SvnChangedFile(filePath, status));
  //    }

    if (status.contains(BazaarStatusKind.UNVERSIONED) || status.contains(BazaarStatusKind.UNKNOWN)) {
      if (vFile != null) {
        if ( LOG.isDebugEnabled() ) LOG.debug( String.format("Change builder <- '%s' as unversioned", path ) );
        recursivelyAddFiles( myUnversioned, vFile );
      }
    }
    else if ( status.contains(BazaarStatusKind.IGNORED) ) {
      if ( LOG.isDebugEnabled() ) LOG.debug( String.format("Change builder <- '%s' as ignored", path ) );
      recursivelyAddFiles( myIgnored, vFile );
    } else {
      Change c = GitChangeUtils.changeForStatus( myProject, myVcsRoot, thisRevision, parentRevision, status,
                                                 previousPath != null ? previousPath.getOriginalPath() : path,
                                                 previousPath != null && !previousPath.isCopy() );
      if ( c != null ) {
        myChanges.add( c );
      }
    }
  }

  private static class PathModification {
    String originalPath;
    boolean copy;

    private PathModification(String originalPath, boolean copy) {
      this.originalPath = originalPath;
      this.copy = copy;
    }

    PathModification forChild(String childRelPath) {
      return new PathModification( originalPath + "/" + childRelPath, copy );
    }

    public String getOriginalPath() {
      return originalPath;
    }

    public void setOriginalPath(String originalPath) {
      this.originalPath = originalPath;
    }

    public boolean isCopy() {
      return copy;
    }

    public void setCopy(boolean copy) {
      this.copy = copy;
    }
  }

  public PathModification addPathModification(final IBazaarStatus status, final PathModification previousPath) {
    final PathModification pathMod;
    if ( status.contains( BazaarStatusKind.RENAMED ) ) {
      pathMod = new PathModification( status.getPreviousPath(), false );
    } else if ( previousPath != null ) {
      pathMod = previousPath;
    } else {
      return null;
    }
    final List<VirtualFile> modifiedFiles = pathMod.isCopy() ? myCopied : myRenamed;
    modifiedFiles.add( myVcsRoot.findFileByRelativePath( status.getPath() ) );

    return myPrevious.put( status.getPath(), pathMod );
  }

  public PathModification addPathModification(final IBazaarStatus status ) {
    return addPathModification( status, null );
  }

  public PathModification getExistingPathModification( String path ) {
    FilePath filePath = VcsUtil.getFilePath( path );
    while ( filePath != null ) {
      PathModification pathMod = myPrevious.get( filePath.getPath() );
      if ( pathMod != null ) {
        return pathMod.forChild( path.substring(filePath.getPath().length() + 1) );
      }
      filePath = filePath.getParentPath();
    }
    return null;
  }

}
