/*
 * Copyright 2000-2008 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package git4idea.checkin;

import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.DialogWrapper;
import com.intellij.openapi.ui.TextComponentAccessor;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.fileChooser.FileChooserDescriptorFactory;
import com.intellij.ui.DocumentAdapter;
/* bzr4idea
import git4idea.GitBranch;
*/
import git4idea.GitRemote;
import git4idea.i18n.GitBundle;
import git4idea.ui.GitUIUtil;
import org.jetbrains.annotations.Nullable;
import org.vcs.bazaar.client.commandline.commands.options.Option;
import org.vcs.bazaar.client.commandline.syntax.IPushOptions;
import org.vcs.bazaar.client.core.BranchLocation;

import javax.swing.*;
import javax.swing.event.DocumentEvent;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;
import java.util.Collection;

/**
 * The push dialog
 */
public class GitPushDialog extends DialogWrapper {
  /**
   * the logger
   */
  private static final Logger LOG = Logger.getInstance(GitPushDialog.class.getName());
  /**
   * Push policy meaning selected references
   */
  private static final String PUSH_POLICY_SELECTED = GitBundle.getString("push.policy.selected");
  /**
   * This push policy means that simple "git push" will be used, so push will happen according to the push configuration.
   */
  private static final String PUSH_POLICY_DEFAULT = GitBundle.getString("push.policy.default");
  /**
   * Push policy meanding all references
   */
  private static final String PUSH_POLICY_ALL = GitBundle.getString("push.policy.all");
  /**
   * Push policy meaning mirror
   */
  private static final String PUSH_POLICY_MIRROR = GitBundle.getString("push.policy.mirror");
  /**
   * The root panel
   */
  private JPanel myPanel;
  /**
   * Git root selector
   */
  private JComboBox myGitRootComboBox;
  /**
   * Remote name combobox
   */
  private com.intellij.ui.ComboboxWithBrowseButton myRemoteComboBox;
  /**
   * Push tags flag
   */
  /* bzr4idea
  private JCheckBox myPushTagsCheckBox;
  */
  /**
   * Use thin pack flag
   */
  /* bzr4idea
  private JCheckBox myUseThinPackCheckBox;
  */
  /**
   * The push policy drop down
   */
  /* bzr4idea
  private JComboBox myPushPolicy;
  */
  /**
   * Force update checkbox
   */
  private JCheckBox myForceUpdateCheckBox;
  private JCheckBox myRememberDefaultCheckbox;
  private JCheckBox myUseExistingCheckBox;
  private JCheckBox myCreatePrefixCheckBox;
  /**
   * Chooser for branches
   */
  /* bzr4idea
  private ElementsChooser<String> myBranchChooser;
  */
  /**
   * The current branch label
   */
  /* bzr4idea
  private JLabel myCurrentBranch;
  */
  /**
   * The checkbox that specifies whether tags are shown
   */
  /* bzr4idea
  private JCheckBox myShowTagsCheckBox;
  */
  /**
   * The list of branches
   */
  /* bzr4idea
  private final ArrayList<String> myBranchNames = new ArrayList<String>();
  */
  /**
   * The list of branches
   */
  /* bzr4idea
  private final ArrayList<String> myTagNames = new ArrayList<String>();
  */
  /**
   * The map from list of branches to the result of mirror checks
   */
  /* bzr4idea
  private final HashMap<String, Boolean> myMirrorChecks = new HashMap<String, Boolean>();
  */
  /**
   * The current project
   */
  private final Project myProject;

  /**
   * A constructor
   *
   * @param project     the project
   * @param roots       the list of the roots
   * @param defaultRoot the default root to select
   */
  public GitPushDialog(final Project project, final List<VirtualFile> roots, final VirtualFile defaultRoot) {
    super(project, true);
    setTitle(GitBundle.getString("push.title"));
    GitUIUtil.setupRootChooser(project, roots, defaultRoot, myGitRootComboBox, null);
    myProject = project;
    setupRemotes();
    setupPolicy();
    setupValidation();
    myRemoteComboBox.getComboBox().setEditable(true);
    myRemoteComboBox.addBrowseFolderListener( GitBundle.getString("common.branch"),
            GitBundle.getString("push.branch.tooltip"), project, FileChooserDescriptorFactory.createSingleFolderDescriptor(),
            TextComponentAccessor.STRING_COMBOBOX_WHOLE_TEXT
    );
    /* bzr4idea
    myShowTagsCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(final ActionEvent e) {
        if (myPushPolicy.getSelectedItem().equals(PUSH_POLICY_SELECTED)) {
          updateBranchChooser();
        }
      }
    });
    */
    init();
  }

  /**
   * @return a prepared handler for push operation
   */
  /* bzr4idea - use getPushOptions()
  public GitLineHandler handler() {
    GitLineHandler h = new GitLineHandler(myProject, getGitRoot(), GitHandler.PUSH);
    String policy = (String)myPushPolicy.getSelectedItem();
    if (PUSH_POLICY_ALL.equals(policy)) {
      h.addParameters("--all");
    }
    else if (PUSH_POLICY_MIRROR.equals(policy)) {
      h.addParameters("--mirror");
    }
    if (myPushTagsCheckBox.isEnabled() && myPushTagsCheckBox.isSelected()) {
      h.addParameters("--tags");
    }
    if (myUseThinPackCheckBox.isSelected()) {
      h.addParameters("--thin");
    }
    if (myForceUpdateCheckBox.isSelected()) {
      h.addParameters("--force");
    }
    h.addParameters("-v");
    h.addParameters(getRemoteTextField().getText());
    if (PUSH_POLICY_SELECTED.equals(policy)) {
      for (String b : myBranchChooser.getMarkedElements()) {
        if (myBranchNames.contains(b)) {
          h.addParameters(b);
        }
        else {
          h.addParameters("tag", b);
        }
      }
    }
    return h;
  }
  */

  public BranchLocation getDestinationBranchLocation() {
    return new BranchLocation( getRemoteTextField().getText() );
  }

  public Collection<Option> getPushOptions() {
    List<Option> options = new ArrayList<Option>();
    if ( myRememberDefaultCheckbox.isSelected() ) {
      options.add( IPushOptions.REMEMBER );
    }
    if ( myCreatePrefixCheckBox.isSelected() ) {
      options.add( IPushOptions.CREATE_PREFIX );
    }
    if ( myUseExistingCheckBox.isSelected() ) {
      options.add( IPushOptions.USE_EXISTING_DIR );
    }
    if ( myForceUpdateCheckBox.isSelected() ) {
      options.add( IPushOptions.OVERWRITE );
    }
    return options;
  }

  /**
   * Setup dialog validation
   */
  private void setupValidation() {
    /* bzr4idea
    myPushPolicy.addActionListener(new ActionListener() {
      public void actionPerformed(final ActionEvent e) {
        validateFields();
      }
    });
    */
    getRemoteTextField().getDocument().addDocumentListener(new DocumentAdapter() {
      protected void textChanged(final DocumentEvent e) {
        validateFields();
      }
    });
    /* bzr4idea
    myBranchChooser.addElementsMarkListener(new ElementsChooser.ElementsMarkListener<String>() {
      public void elementMarkChanged(final String element, final boolean isMarked) {
        validateFields();
      }
    });
    */
    validateFields();
  }

  /**
   * Validate fields in the dialog
   */
  private void validateFields() {
    boolean isValid = getRemoteTextField().getText().trim().length() != 0;
    /* bzr4idea
    final Object policy = myPushPolicy.getSelectedItem();
    isValid &= !policy.equals(PUSH_POLICY_SELECTED) || myBranchChooser.getMarkedElements().size() != 0;
    */
    setOKActionEnabled(isValid);
  }

  /**
   * Setup policy combobox
   */
  private void setupPolicy() {
    /* bzr4idea
    myPushPolicy.addItem(PUSH_POLICY_SELECTED);
    myPushPolicy.addItem(PUSH_POLICY_DEFAULT);
    myPushPolicy.addItem(PUSH_POLICY_ALL);
    myPushPolicy.addItem(PUSH_POLICY_MIRROR);
    myPushPolicy.setSelectedIndex(0);
    // configure policy listner
    final ActionListener policyListener = new ActionListener() {
      public void actionPerformed(final ActionEvent e) {
        String p = (String)myPushPolicy.getSelectedItem();
        if (PUSH_POLICY_SELECTED.equals(p)) {
          myBranchChooser.setEnabled(true);
          myShowTagsCheckBox.setEnabled(true);
          updateBranchChooser();
        }
        else {
          myBranchChooser.clear();
          myBranchChooser.setEnabled(false);
          myShowTagsCheckBox.setEnabled(false);
        }
        if (PUSH_POLICY_MIRROR.equals(p)) {
          myPushTagsCheckBox.setEnabled(false);
          myPushTagsCheckBox.setSelected(true);
          myForceUpdateCheckBox.setEnabled(false);
          myForceUpdateCheckBox.setSelected(true);
        }
        else {
          if (!myForceUpdateCheckBox.isEnabled()) {
            myForceUpdateCheckBox.setEnabled(true);
            myForceUpdateCheckBox.setSelected(false);
          }
          if (PUSH_POLICY_ALL.equals(p)) {
            myPushTagsCheckBox.setEnabled(false);
            myPushTagsCheckBox.setSelected(false);
          }
          else if (!myPushTagsCheckBox.isEnabled()) {
            myPushTagsCheckBox.setEnabled(true);
            myPushTagsCheckBox.setSelected(false);
          }
        }
      }
    };
    myPushPolicy.addActionListener(policyListener);
    policyListener.actionPerformed(null);
    */
    // select remote listener
    final DocumentAdapter listener = new DocumentAdapter() {
      VirtualFile myPreviousRoot;
      GitRemote myPreviousRemote = null;

      protected void textChanged(final DocumentEvent e) {
        final VirtualFile newRoot = getGitRoot();
        final GitRemote newRemote = getRemote(getRemoteTextField().getText());
        if (newRoot == null) {
          return;
        }
        if (myPreviousRoot == null || myPreviousRemote == null || !myPreviousRoot.equals(newRoot) || !myPreviousRemote.equals(newRemote)) {
          /* bzr4idea
          if (isMirror()) {
            myPushPolicy.setEnabled(false);
            myPushPolicy.setSelectedItem(PUSH_POLICY_MIRROR);
          }
          else {
            if (!myPushPolicy.isEnabled()) {
              myPushPolicy.setSelectedItem(PUSH_POLICY_SELECTED);
              myPushPolicy.setEnabled(true);
            }
          }
          */
          myPreviousRoot = newRoot;
          myPreviousRemote = newRemote;
        }
      }
    };
    getRemoteTextField().getDocument().addDocumentListener(listener);
    myGitRootComboBox.addActionListener(new ActionListener() {
      public void actionPerformed(final ActionEvent e) {
        listener.changedUpdate(null);
      }
    });
    listener.changedUpdate(null);
  }

  /**
   * Update content of the branch chooser
   */
  /* bzr4idea
  private void updateBranchChooser() {
    myBranchChooser.clear();
    final String current = myCurrentBranch.getText();
    for (String b : myBranchNames) {
      myBranchChooser.addElement(b, b.equals(current));
    }
    if (myShowTagsCheckBox.isSelected()) {
      for (String t : myTagNames) {
        myBranchChooser.addElement(t, false);
      }
    }
    validateFields();
  }
  */

  /**
   * @return true if the current branch should be mirror branch
   */
  /* bzr4idea
  private boolean isMirror() {
    final String name = getRemoteTextField().getText();
    Boolean rc = myMirrorChecks.get(name);
    if (rc == null) {
      rc = false;
      GitRemote remote = getRemote(name);
      if (remote != null) {
        try {
          rc = GitConfigUtil.getBoolValue(myProject, getGitRoot(), "remote." + name + ".mirror");
          if (rc == null) {
            rc = false;
          }
        }
        catch (VcsException e) {
          // treat error as a false value
        }
        myMirrorChecks.put(name, rc);
      }
    }
    return rc.booleanValue();
  }
  */

  /**
   * Get currently selected remote object.
   *
   * @param name a name to select
   * @return the remote or null
   */
  @Nullable
  private GitRemote getRemote(final String name) {
    for (int i = myRemoteComboBox.getComboBox().getItemCount() - 1; i >= 0; i--) {
      final GitRemote r = (GitRemote)myRemoteComboBox.getComboBox().getItemAt(i);
      if (name.equals(r.toString())) {
        return r;
      }
    }
    return null;
  }

  /**
   * Setup drop down with remotes
   */
  private void setupRemotes() {
    final ActionListener actionListener = new ActionListener() {
      public void actionPerformed(final ActionEvent e) {
        updateRemotes();
        /* bzr4idea - braches are disabled - how they differ from remotes?
                      tags are disabled - not implemented by bzr-java-lib
        myMirrorChecks.clear();
        myBranchNames.clear();
        myTagNames.clear();
        try {
          GitBranch.listAsStrings(myProject, getGitRoot(), false, true, myBranchNames);
          GitTag.listAsStrings(myProject, getGitRoot(), myTagNames);
        }
        catch (VcsException ex) {
          LOG.warn("Exception in branchlist: \n" + StringUtil.getThrowableText(ex));
        }
        */
      }
    };
    myGitRootComboBox.addActionListener(actionListener);
    actionListener.actionPerformed(null);
  }

  /**
   * Update remotes
   */
  private void updateRemotes() {
    GitUIUtil.setupRemotes(myProject, getGitRoot(), myRemoteComboBox.getComboBox());
  }

  /**
   * @return text field for {@link #myRemoteComboBox}
   */
  private JTextField getRemoteTextField() {
    return (JTextField)myRemoteComboBox.getComboBox().getEditor().getEditorComponent();
  }

  /**
   * @return the currently selected git root
   */
  private VirtualFile getGitRoot() {
    return (VirtualFile)myGitRootComboBox.getSelectedItem();
  }

  /**
   * {@inheritDoc}
   */
  protected JComponent createCenterPanel() {
    return myPanel;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  protected String getDimensionServiceKey() {
    return getClass().getName();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  protected String getHelpId() {
    return "reference.VersionControl.Bzr.Push";
  }

  /**
   * Create UI components
   */
  private void createUIComponents() {
    /* bzr4idea
    myBranchChooser = new ElementsChooser<String>(true);
    */
  }
}
