package git4idea.config;
/*
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for
 * the specific language governing permissions and limitations under the License.
 *
 * Copyright 2007 Decentrix Inc
 * Copyright 2007 Aspiro AS
 * Copyright 2008 MQSoftware
 * Copyright 2008 JetBrains s.r.o.
 * 
 * Authors: gevession, Erlend Simonsen & Mark Scott
 *
 * This code was originally derived from the MKS & Mercurial IDEA VCS plugins
 */

import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.components.PersistentStateComponent;
import com.intellij.openapi.components.ServiceManager;
import com.intellij.openapi.components.State;
import com.intellij.openapi.components.Storage;
import com.intellij.openapi.progress.ProcessCanceledException;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Computable;
import com.intellij.openapi.util.SystemInfo;
import com.intellij.util.xmlb.XmlSerializerUtil;
import org.jetbrains.annotations.NonNls;

import java.io.File;
import java.util.Arrays;
import java.util.LinkedList;

/**
 * Git VCS settings
 */
@State(
  name = "Bazaar.Settings",
  storages = {@Storage(
    id = "ws",
    file = "$WORKSPACE_FILE$")})
public class BzrVcsSettings implements PersistentStateComponent<BzrVcsSettings> {
  /**
   * the default cygwin executable
   */
  @NonNls private static final String[] DEFAULT_WINDOWS_PATHS =
    {"C:\\cygwin\\bin", "C:\\Program Files\\Bazaar"};
  /**
   * Windows executable name
   */
  @NonNls private static final String DEFAULT_WINDOWS_GIT = "bzr.exe";
  /**
   * Default unix paths
   */
  @NonNls private static final String[] DEFAULT_UNIX_PATHS = {"/usr/local/bin", "/usr/bin", "/opt/local/bin", "/opt/bin"};
  /**
   * Unix executable name
   */
  @NonNls private static final String DEFAULT_UNIX_GIT = "bzr";

  /**
   * The default executable for GIT
   */
  public String GIT_EXECUTABLE = defaultGit();
  /**
   * The previously entered authors of the commit (up to 10)
   */
  public String[] PREVIOUS_COMMIT_AUTHORS = {};
  /**
   * Limit for previous commit authors
   */
  private static final int PREVIOUS_COMMIT_AUTHORS_LIMIT = 16;
  /**
   * Checkout includes tags
   */
  public Boolean CHECKOUT_INCLUDE_TAGS;

  public boolean cygwin;

  /**
   * Save an author of the commit and make it the first one. If amount of authors exceeds the limit, remove least recently selected author.
   *
   * @param author an author to save
   */
  public void saveCommitAuthor(String author) {
    LinkedList<String> authors = new LinkedList<String>(Arrays.asList(PREVIOUS_COMMIT_AUTHORS));
    authors.remove(author);
    while (authors.size() >= PREVIOUS_COMMIT_AUTHORS_LIMIT) {
      authors.removeLast();
    }
    authors.addFirst(author);
    PREVIOUS_COMMIT_AUTHORS = authors.toArray(new String[authors.size()]);
  }

  /**
   * {@inheritDoc}
   */
  public BzrVcsSettings getState() {
    return this;
  }

  /**
   * {@inheritDoc}
   */
  public void loadState(BzrVcsSettings bzrVcsSettings) {
    XmlSerializerUtil.copyBean(bzrVcsSettings, this);
  }

  /**
   * Get git setting for the project
   *
   * @param project a context project
   * @return the git settigns
   */
  public static BzrVcsSettings getInstance(Project project) {
    return ServiceManager.getService(project, BzrVcsSettings.class);
  }

  /**
   * Get instance with checked read action
   *
   * @param project the project to get setting for
   * @return the settings object
   */
  public static BzrVcsSettings getInstanceChecked(final Project project) {
    return ApplicationManager.getApplication().runReadAction(new Computable<BzrVcsSettings>() {
      public BzrVcsSettings compute() {
        if (project.isDisposed()) throw new ProcessCanceledException();
        return ServiceManager.getService(project, BzrVcsSettings.class);
      }
    });
  }

  /**
   * @return the default executable name depending on the platform
   */
  private static String defaultGit() {
    String[] paths;
    String exe;
    if (SystemInfo.isWindows) {
      exe = DEFAULT_WINDOWS_GIT;
      paths = DEFAULT_WINDOWS_PATHS;
    }
    else {
      exe = DEFAULT_UNIX_GIT;
      paths = DEFAULT_UNIX_PATHS;
    }
    for (String p : paths) {
      File f = new File(p, exe);
      if (f.exists()) {
        return f.getAbsolutePath();
      }
    }
    return exe;     // otherwise, hope it's in $PATH
  }
}
