package git4idea.config;
/*
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for
 * the specific language governing permissions and limitations under the License.
 *
 * Copyright 2007 Decentrix Inc
 * Copyright 2007 Aspiro AS
 * Copyright 2008 MQSoftware
 * Authors: gevession, Erlend Simonsen & Mark Scott
 *
 * This code was originally derived from the MKS & Mercurial IDEA VCS plugins
 */

import com.intellij.openapi.fileChooser.FileChooserDescriptor;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.ui.TextFieldWithBrowseButton;
import com.intellij.openapi.vcs.VcsException;
import com.intellij.openapi.options.ConfigurationException;
import git4idea.GitVcs;
import git4idea.i18n.GitBundle;
import org.jetbrains.annotations.NotNull;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import bzr4idea.Bzr;
import bzr4idea.BzrException;
import bzr4idea.util.Log;

/**
 * Git VCS configuration panel
 */
public class GitVcsPanel {
  private JButton myTestButton;
  private JComponent myPanel;
  private TextFieldWithBrowseButton myGitField;
  private JCheckBox cygwinCheckBox;
  private final Project myProject;
  private final BzrVcsSettings mySettings;

  public GitVcsPanel(@NotNull Project project) {
    mySettings = BzrVcsSettings.getInstance(project);
    myProject = project;
    myTestButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        testConnection();
      }
    });

    myGitField.addBrowseFolderListener(GitBundle.getString("find.git.title"), GitBundle.getString("find.git.description"), project,
                                       new FileChooserDescriptor(true, false, false, false, false, false));
  }

  private void testConnection() {
    boolean modified = isModified(mySettings);
    final String s;
    try {
      if (modified) {
        BzrVcsSettings tmpsettings = new BzrVcsSettings();
        tmpsettings.GIT_EXECUTABLE = myGitField.getText();
        Bzr.initBazaarClient(tmpsettings);
      }
      s = GitVcs.version(myProject);
    }
    catch (VcsException e) {
      Messages.showErrorDialog(myProject, e.getMessage(), GitBundle.getString("find.git.error.title"));
      return;
    } finally {
      if (modified) {
        try {
          Bzr.initBazaarClient(mySettings);
        } catch (BzrException e) {
          Log.error(e, "Unable to restore previous setup");
        }
      }
    }
    if (GitVersion.parse(s).isSupported()) {
      Messages.showInfoMessage(myProject, GitBundle.message("find.git.success.message", s), GitBundle.getString("find.git.success.title"));
    }
    else {
      Messages.showWarningDialog(myProject, GitBundle.message("find.git.unsupported.message", s, GitVersion.MIN),
                                 GitBundle.getString("find.git.success.title"));
    }
  }

  public JComponent getPanel() {
    return myPanel;
  }

  public void load(@NotNull BzrVcsSettings settings) {
    myGitField.setText(settings.GIT_EXECUTABLE);
    cygwinCheckBox.setSelected(settings.cygwin);
    if (!Bzr.isWindows()) {
      cygwinCheckBox.setVisible(false);
    }
  }

  public boolean isModified(@NotNull BzrVcsSettings settings) {
    return !settings.GIT_EXECUTABLE.equals(myGitField.getText())
        || settings.cygwin != cygwinCheckBox.isSelected();
  }

  public void save(@NotNull BzrVcsSettings settings) throws ConfigurationException {
    try {
      BzrVcsSettings newsettings = new BzrVcsSettings();
      newsettings.GIT_EXECUTABLE = myGitField.getText();
      newsettings.cygwin = cygwinCheckBox.isSelected();
      Bzr.initBazaarClient(newsettings);
    } catch (BzrException e) {
      throw new ConfigurationException(e.toString());
    }
    settings.GIT_EXECUTABLE = myGitField.getText();
    settings.cygwin = cygwinCheckBox.isSelected();
  }
}
