package git4idea.history;
/*
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for
 * the specific language governing permissions and limitations under the License.
 *
 * Copyright 2008 MQSoftware
 * Authors: Mark Scott
 *
 * This code was originally derived from the MKS IDEA VCS plugin
 */

import com.intellij.openapi.vcs.history.HistoryAsTreeProvider;
import com.intellij.openapi.vcs.history.VcsFileRevision;
import com.intellij.openapi.vcs.history.VcsRevisionNumber;
import com.intellij.util.TreeItem;
import git4idea.GitRevisionNumber;
import git4idea.GitFileRevision;

import java.util.*;

public class GitHistoryTreeProvider implements HistoryAsTreeProvider {
  private static void addMergedRevisions( Map<VcsRevisionNumber, TreeItem<VcsFileRevision>> treeItemMap, TreeItem<VcsFileRevision> node, GitFileRevision revision ) {
    List<GitFileRevision> mergedRevs = revision.getMergedRevisions();
    if ( mergedRevs == null ) return;
    
    try {
      for (GitFileRevision mergeRev : mergedRevs) {
        final TreeItem<VcsFileRevision> childItem;
        childItem = new TreeItem<VcsFileRevision>(mergeRev);
        node.addChild( childItem );
        treeItemMap.put(mergeRev.getRevisionNumber(), childItem);
        addMergedRevisions( treeItemMap, childItem, mergeRev );
      }
    }
    catch (Exception e) {
    }
  }

  private static void processParentRevisions( List<TreeItem<VcsFileRevision>> result, Map<VcsRevisionNumber, TreeItem<VcsFileRevision>> treeItemMap, TreeItem<VcsFileRevision> node, GitFileRevision revision ) {
    // now look for parents and set parent/child relationships
    List<String> parentRevStrings = revision.getParentRevisions();

    try {
      for ( String parentRevString : parentRevStrings ) {
        if (parentRevString != null && !"".equals(parentRevString)) {
          final TreeItem<VcsFileRevision> parentItem;
          parentItem = treeItemMap.get(GitRevisionNumber.createRevision(parentRevString));
          if (parentItem != null) {
//                                      parentItem.addChild(treeItem);
            // we want to keep newer revisions on top, thus reverse the order
            parentItem.addChild(node);
//                                              treeItem.addChild(parentItem);
            // remove children so they don't appear multiple times
            if (result.contains(node)) {
              result.remove(node);
            }
          }
        }
      }
    }
    catch (Exception e) {
    }
  }

  /**
   * It seems we should only return roots
   *
   * @param allRevisions all the revisions to be viewed as a tree (will become node/treeitems)
   * @return a list of the roots, usually the only root (aka revision 1.1)
   */
  @SuppressWarnings({"EmptyCatchBlock"})
  public List<TreeItem<VcsFileRevision>> createTreeOn(List<VcsFileRevision> allRevisions) {
    Map<VcsRevisionNumber, TreeItem<VcsFileRevision>> treeItemMap = new HashMap<VcsRevisionNumber, TreeItem<VcsFileRevision>>();
    /* bzr4idea - no reverse sorting
    Map<VcsRevisionNumber, VcsFileRevision> revisionsByRevNumber = new HashMap<VcsRevisionNumber, VcsFileRevision>();
    for (VcsFileRevision revision : allRevisions) {
      revisionsByRevNumber.put(revision.getRevisionNumber(), revision);
    }
    // first order the revisions, so we can simply process the list and be sure
    // parent revisions have been processed first
    List<VcsRevisionNumber> orderedRevisions = new ArrayList<VcsRevisionNumber>(revisionsByRevNumber.keySet());
    Collections.sort(orderedRevisions);
    */
    List<TreeItem<VcsFileRevision>> result = new ArrayList<TreeItem<VcsFileRevision>>(allRevisions.size());

    for (VcsFileRevision revision : allRevisions) {
      TreeItem<VcsFileRevision> treeItem = new TreeItem<VcsFileRevision>(revision);
      treeItemMap.put(revision.getRevisionNumber(), treeItem);
      result.add(treeItem);

      // append merged revisions
      addMergedRevisions( treeItemMap, treeItem, (GitFileRevision)revision );
    }

    // rebuild tree with respect to parents data
    for ( TreeItem<VcsFileRevision> treeItem : treeItemMap.values() ) {
      processParentRevisions( result, treeItemMap, treeItem, (GitFileRevision)treeItem.getData() );
    }

    return result;
  }
}
