/*
 * Copyright 2000-2008 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package git4idea.history;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.vcs.FilePath;
import com.intellij.openapi.vcs.VcsException;
import com.intellij.openapi.vcs.changes.Change;
import com.intellij.openapi.vcs.changes.ChangeListManager;
import com.intellij.openapi.vcs.history.VcsFileRevision;
import com.intellij.openapi.vcs.history.VcsRevisionNumber;
import com.intellij.openapi.vfs.VirtualFile;
import git4idea.GitContentRevision;
import git4idea.GitFileRevision;
import git4idea.GitRevisionNumber;
import git4idea.GitUtil;
import git4idea.i18n.GitBundle;
import git4idea.commands.GitHandlerUtil;
/* git4idea
import git4idea.commands.GitHandler;
import git4idea.commands.GitSimpleHandler;
*/
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.NotNull;
import org.vcs.bazaar.client.IBazaarLogMessage;
import org.vcs.bazaar.client.BazaarRevisionRange;
import org.vcs.bazaar.client.BazaarRevision;

import java.util.ArrayList;
import java.util.List;

import bzr4idea.Bzr;
import bzr4idea.BzrException;

/**
 * History utilities for Git
 */
public class GitHistoryUtils {
  /**
   * A private constructor
   */
  private GitHistoryUtils() {
  }

  /**
   * Get current revision for the file under git
   *
   * @param project  a project
   * @param filePath a file path
   * @return a revision number or null if the file is unversioned or new
   * @throws VcsException if there is problem with running git
   */
  @Nullable
  public static VcsRevisionNumber getCurrentRevision(final Project project, FilePath filePath) throws VcsException {
    final VirtualFile root = GitUtil.getGitRoot(project, filePath);
    /* git4idea
    filePath = getLastCommitName(project, filePath);
    GitSimpleHandler h = new GitSimpleHandler(project, GitUtil.getGitRoot(project, filePath), GitHandler.LOG);
    h.setNoSSH(true);
    h.setSilent(true);
    h.addParameters("-n1", "--pretty=format:%H%n%ct%n");
    h.endOptions();
    h.addRelativePaths(filePath);
    String result = h.run();
    */
    Bzr bzr = Bzr.forBzrRoot( root );
    List<IBazaarLogMessage> output = bzr.log( filePath, null, 1 );

    /* git4idea
    if (result.length() == 0) {
      return null;
    }
    String[] lines = result.split("\n");
    String revstr = lines[0];
    Date commitDate = GitUtil.parseTimestamp(lines[1]);
    return new GitRevisionNumber(revstr, commitDate);
    */
    if ( output != null && !output.isEmpty() ) {
      return new GitRevisionNumber( output.get(0).getRevision(), output.get(0).getDate() );
    } else {
      return null;
    }
  }


  /**
   * Get history for the file
   *
   * @param project the context project
   * @param path    the file path
   * @return the list of the revisions
   * @throws VcsException if there is problem with running git
   */
  public static List<VcsFileRevision> history(final Project project, FilePath path) throws VcsException {
    /* git4idea
    // adjust path using change manager
    */
    final FilePath adjPath = getLastCommitName(project, path);
    final VirtualFile root = GitUtil.getGitRoot(project, path);
    /* git4idea
    GitSimpleHandler h = new GitSimpleHandler(project, root, GitHandler.LOG);
    h.setNoSSH(true);
    h.setStdoutSuppressed(true);
    h.addParameters("-M", "--follow", "--name-only", "--pretty=format:%H%x00%ct%x00%an%x20%x3C%ae%x3E%x00%cn%x20%x3C%ce%x3E%x00%s%n%n%b%x00",
                    "--encoding=UTF-8");
    h.endOptions();
    h.addRelativePaths(path);
    String output = h.run();
    */

    final List<VcsFileRevision> rc = new ArrayList<VcsFileRevision>();
    GitHandlerUtil.doSynchronously(
        new GitHandlerUtil.BzrSynchronousRunner( GitBundle.message("getting.history", path.getPresentableUrl() ),
                                                 Bzr.forBzrRoot( root ), project ){
          protected void run(Bzr bzr) throws BzrException {
            final List<IBazaarLogMessage> output = bzr.log( adjPath, null, null );
            appendBranchLog( project, adjPath, rc, output );
          }
        }, false, true
    );

    /* git4idea
    StringTokenizer tk = new StringTokenizer(output, "\u0000\n", false);
    String prefix = root.getPath() + "/";
    while (tk.hasMoreTokens()) {
      final GitRevisionNumber revision = new GitRevisionNumber(tk.nextToken("\u0000\n"), GitUtil.parseTimestamp(tk.nextToken("\u0000")));
      final String author = GitUtil.adjustAuthorName(tk.nextToken("\u0000"), tk.nextToken("\u0000"));
      final String message = tk.nextToken("\u0000").trim();
      final FilePath revisionPath = VcsUtil.getFilePathForDeletedFile(prefix + GitUtil.unescapePath(tk.nextToken("\u0000\n")), false);
      rc.add(new GitFileRevision(project, revisionPath, revision, author, message, null));
    }
    */
    return rc;
  }

  public static void appendBranchLog( @NotNull Project project, @NotNull FilePath logPath, @NotNull List<VcsFileRevision> revList, List<IBazaarLogMessage> messages )
  {
    for ( IBazaarLogMessage logEntry : messages ) {
      revList.add( processLogEntry( project, logPath, logEntry ) );
      if (!logEntry.getMerged().isEmpty()) {
        appendBranchLog( project, logPath, revList, logEntry.getMerged() );
      }
    }
  }

  public static GitFileRevision processLogEntry(@NotNull Project project, @NotNull FilePath logPath, IBazaarLogMessage logEntry) {
    final List<GitFileRevision> mergedRevisions;
    if ( !logEntry.getMerged().isEmpty() ) {
      mergedRevisions = new ArrayList<GitFileRevision>();
      for ( IBazaarLogMessage mergedEntry : logEntry.getMerged() ) {
        mergedRevisions.add( processLogEntry( project, logPath, mergedEntry ) );
      }
    } else {
      mergedRevisions = null;
    }
    return new GitFileRevision( project, logPath, new GitRevisionNumber( logEntry.getRevision(), logEntry.getDate() ),
                         logEntry.getAuthor(), logEntry.getMessage(), logEntry.getBranchNick(), logEntry.getParents(), mergedRevisions );
  }

  /**
   * Get name of the file in the last commit. If file was renamed, returns the previous name.
   *
   * @param project the contenxt project
   * @param path    the path to check
   * @return the name of file in the last commit or argument
   */
  public static FilePath getLastCommitName(final Project project, FilePath path) {
    final ChangeListManager changeManager = ChangeListManager.getInstance(project);
    final Change change = changeManager.getChange(path);
    if (change != null && change.getType() == Change.Type.MOVED) {
      GitContentRevision r = (GitContentRevision)change.getBeforeRevision();
      assert r != null : "Move change always have beforeRevision";
      path = r.getFile();
    }
    return path;
  }
}
