/*
 * Copyright 2000-2008 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package git4idea.vfs;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.vcs.FilePath;
import com.intellij.openapi.vcs.VcsException;
import com.intellij.openapi.vcs.VcsVFSListener;
import com.intellij.openapi.vcs.changes.VcsDirtyScopeManager;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.vcsUtil.VcsUtil;
import static com.intellij.vcsUtil.VcsUtil.getVirtualFile;
import git4idea.GitUtil;
import git4idea.GitVcs;
import git4idea.commands.GitFileUtils;
import git4idea.i18n.GitBundle;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import bzr4idea.Bzr;
import bzr4idea.util.Log;

/**
 * Git virtual file adapter
 */
public class GitVFSListener extends VcsVFSListener {
  /**
   * Dirty scope manager for the project
   */
  private final VcsDirtyScopeManager myDirtyScopeManager;

  /**
   * A constructor for listener
   *
   * @param project a project
   * @param vcs     a vcs for that proejct
   */
  public GitVFSListener(final Project project, final GitVcs vcs) {
    super(project, vcs);
    myDirtyScopeManager = VcsDirtyScopeManager.getInstance(myProject);
  }

  /**
   * {@inheritDoc}
   */
  protected String getAddTitle() {
    return GitBundle.getString("vfs.listener.add.title");
  }

  /**
   * {@inheritDoc}
   */
  protected String getSingleFileAddTitle() {
    return GitBundle.getString("vfs.listener.add.single.title");
  }

  /**
   * {@inheritDoc}
   */
  protected String getSingleFileAddPromptTemplate() {
    return GitBundle.getString("vfs.listener.add.single.prompt");
  }

  /**
   * {@inheritDoc}
   */
  protected void performAdding(final Collection<VirtualFile> addedFiles, final Map<VirtualFile, VirtualFile> copyFromMap) {
    Map<VirtualFile, List<VirtualFile>> sortedFiles = GitUtil.sortFilesByGitRoot(addedFiles);
    // note that copied files are not processed because they are included into added files.
    for (Map.Entry<VirtualFile, List<VirtualFile>> e : sortedFiles.entrySet()) {
      try {
        final VirtualFile root = e.getKey();
        GitFileUtils.addFiles(myProject, root, e.getValue());
        markRootDirty(root);
      }
      catch (VcsException ex) {
        ((GitVcs)myVcs).showMessages(ex.getMessage());
      }
    }
  }

  /**
   * {@inheritDoc}
   */
  protected String getDeleteTitle() {
    return GitBundle.getString("vfs.listener.delete.title");
  }

  /**
   * {@inheritDoc}
   */
  protected String getSingleFileDeleteTitle() {
    return GitBundle.getString("vfs.listener.delete.single.title");
  }

  /**
   * {@inheritDoc}
   */
  protected String getSingleFileDeletePromptTemplate() {
    return GitBundle.getString("vfs.listener.delete.single.prompt");
  }

  /**
   * {@inheritDoc}
   */
  protected void performDeletion(final List<FilePath> filesToDelete) {
    Map<VirtualFile, List<FilePath>> sortedFiles = GitUtil.sortFilePathsByVcsRoot(myProject, filesToDelete);
    for (Map.Entry<VirtualFile, List<FilePath>> e : sortedFiles.entrySet()) {
      try {
        final VirtualFile root = e.getKey();
        GitFileUtils.delete(myProject, root, e.getValue());
        markRootDirty(root);
      }
      catch (VcsException ex) {
        ((GitVcs)myVcs).showMessages(ex.getMessage());
      }
    }
  }

  /**
   * Mark root as dirty
   *
   * @param root a vcs root to rescan
   */
  private void markRootDirty(final VirtualFile root) {
    // Note that the root is invalidated because changes are detected per-root anyway.
    // Otherwise it is not possible to detect moves.
    myDirtyScopeManager.dirDirtyRecursively(root);
  }

  /**
   * {@inheritDoc}
   */
  protected void performMoveRename(final List<MovedFileInfo> movedFiles) {
/*
    // because git does not tracks moves, the file are just added and deleted.
    ArrayList<VirtualFile> added = new ArrayList<VirtualFile>();
    ArrayList<FilePath> removed = new ArrayList<FilePath>();
    for (MovedFileInfo m : movedFiles) {
      added.add(VcsUtil.getVirtualFile(m.myNewPath));
      removed.add(VcsUtil.getFilePath(m.myOldPath));
    }
    performAdding(added, null);
    performDeletion(removed);
*/
    // todo this might need optimization (sort by root, process roots separately)
    for (MovedFileInfo m : movedFiles) {
      try {
        FilePath from = VcsUtil.getFilePath(m.myOldPath);
        VirtualFile to = getVirtualFile(m.myNewPath);
        VirtualFile root = Bzr.root(to);

        Bzr bzr = Bzr.forBzrRoot(root);
        bzr.move(from, to);

        markRootDirty(root);

      } catch (Exception e) {
        Log.error(e, "Error renaming %s to %s", m.myOldPath, m.myNewPath);
        ((GitVcs)myVcs).showMessages(e.toString());
      }
    }
  }

  /**
   * {@inheritDoc}
   */
  protected boolean isDirectoryVersioningSupported() {
    return true;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  protected Collection<FilePath> selectFilePathsToDelete(final List<FilePath> deletedFiles) {
    // For git asking about vcs delete does not make much sense. The result is practically identical.
    return deletedFiles;
  }
}